@@ -71,7 +71,7 @@ start_cfg(void)
 		cfg_add_cause("%s: %s", TMUX_CONF, strerror(errno));
 
 	if (cfg_file == NULL && (home = find_home()) != NULL) {
-		xasprintf(&cfg_file, "%s/.tmux.conf", home);
+		xasprintf(&cfg_file, "%s/.tmate.conf", home);
 		if (access(cfg_file, R_OK) != 0 && errno == ENOENT) {
 			free(cfg_file);
 			cfg_file = NULL;
@@ -81,20 +81,6 @@ start_cfg(void)
 		cfg_add_cause("%s: %s", cfg_file, cause);
 	free(cause);
 
-#ifdef TMATE
-	cause = NULL;
-	if ((home = find_home()) != NULL) {
-		xasprintf(&tmate_cfg_file, "%s/.tmate.conf", home);
-		if (access(tmate_cfg_file, R_OK) != 0 && errno == ENOENT) {
-			free(tmate_cfg_file);
-			tmate_cfg_file = NULL;
-		}
-	}
-	if (tmate_cfg_file != NULL && load_cfg(tmate_cfg_file, cfg_cmd_q, &cause) == -1)
-		cfg_add_cause("%s: %s", cfg_file, cause);
-	free(cause);
-#endif
-
 	cmdq_continue(cfg_cmd_q);
 }
 
@@ -149,6 +135,20 @@ load_cfg(const char *path, struct cmd_q *cmdq, char **cause)
 	return (found);
 }
 
+static void print_cfg_errors(void)
+{
+	u_int i;
+
+	for (i = 0; i < cfg_ncauses; i++) {
+		tmate_info("%s", cfg_causes[i]);
+		free(cfg_causes[i]);
+	}
+
+	free(cfg_causes);
+	cfg_causes = NULL;
+	cfg_ncauses = 0;
+}
+
 void
 cfg_default_done(__unused struct cmd_q *cmdq)
 {
@@ -157,7 +157,14 @@ cfg_default_done(__unused struct cmd_q *cmdq)
 	cfg_finished = 1;
 
 #ifdef TMATE
+	/* We do it this late, this way, CLI options take precedence over cfg file */
+	tmate_load_cli_options();
+
 	tmate_session_start();
+	if (tmate_foreground && cfg_ncauses) {
+		print_cfg_errors();
+		exit(1);
+	}
 #endif
 
 	if (!RB_EMPTY(&sessions))
