@@ -1,7 +1,7 @@
-/* $Id$ */
+/* $OpenBSD$ */
 
 /*
- * Copyright (c) 2007 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2007 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -17,13 +17,12 @@
  */
 
 #include <sys/types.h>
-#include <sys/ioctl.h>
 
 #include <errno.h>
 #include <fcntl.h>
 #include <fnmatch.h>
-#include <pwd.h>
 #include <signal.h>
+#include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
 #include <termios.h>
@@ -58,11 +57,22 @@ struct windows windows;
 struct window_pane_tree all_window_panes;
 u_int	next_window_pane_id;
 u_int	next_window_id;
+u_int	next_active_point;
 
 void	window_pane_timer_callback(int, short, void *);
 void	window_pane_read_callback(struct bufferevent *, void *);
 void	window_pane_error_callback(struct bufferevent *, short, void *);
 
+struct window_pane *window_pane_choose_best(struct window_pane **, u_int);
+
+RB_GENERATE(windows, window, entry, window_cmp);
+
+int
+window_cmp(struct window *w1, struct window *w2)
+{
+	return (w1->id - w2->id);
+}
+
 RB_GENERATE(winlinks, winlink, entry, winlink_cmp);
 
 int
@@ -244,71 +254,75 @@ winlink_stack_remove(struct winlink_stack *stack, struct winlink *wl)
 	}
 }
 
-int
-window_index(struct window *s, u_int *i)
+struct window *
+window_find_by_id_str(const char *s)
 {
-	for (*i = 0; *i < ARRAY_LENGTH(&windows); (*i)++) {
-		if (s == ARRAY_ITEM(&windows, *i))
-			return (0);
-	}
-	return (-1);
+	const char	*errstr;
+	u_int		 id;
+
+	if (*s != '@')
+		return (NULL);
+
+	id = strtonum(s + 1, 0, UINT_MAX, &errstr);
+	if (errstr != NULL)
+		return (NULL);
+	return (window_find_by_id(id));
 }
 
 struct window *
 window_find_by_id(u_int id)
 {
-	struct window	*w;
-	u_int		 i;
+	struct window	w;
 
-	for (i = 0; i < ARRAY_LENGTH(&windows); i++) {
-		w = ARRAY_ITEM(&windows, i);
-		if (w->id == id)
-			return (w);
-	}
-	return (NULL);
+	w.id = id;
+	return (RB_FIND(windows, &windows, &w));
+}
+
+void
+window_update_activity(struct window *w)
+{
+	gettimeofday(&w->activity_time, NULL);
+	alerts_queue(w, WINDOW_ACTIVITY);
 }
 
 struct window *
 window_create1(u_int sx, u_int sy)
 {
 	struct window	*w;
-	u_int		 i;
 
 	w = xcalloc(1, sizeof *w);
-	w->id = next_window_id++;
 	w->name = NULL;
 	w->flags = 0;
 
 	TAILQ_INIT(&w->panes);
 	w->active = NULL;
 
+#ifdef TMATE
+	w->tmate_last_sync_active_pane = NULL;
+#endif
+
 	w->lastlayout = -1;
 	w->layout_root = NULL;
 
 	w->sx = sx;
 	w->sy = sy;
 
-	options_init(&w->options, &global_w_options);
-	if (options_get_number(&w->options, "automatic-rename"))
-		queue_window_name(w);
+	w->options = options_create(global_w_options);
 
-	for (i = 0; i < ARRAY_LENGTH(&windows); i++) {
-		if (ARRAY_ITEM(&windows, i) == NULL) {
-			ARRAY_SET(&windows, i, w);
-			break;
-		}
-	}
-	if (i == ARRAY_LENGTH(&windows))
-		ARRAY_ADD(&windows, w);
 	w->references = 0;
 
+	w->id = next_window_id++;
+	RB_INSERT(windows, &windows, w);
+
+	window_update_activity(w);
+
 	return (w);
 }
 
 struct window *
-window_create(const char *name, const char *cmd, const char *shell,
-    const char *cwd, struct environ *env, struct termios *tio,
-    u_int sx, u_int sy, u_int hlimit, char **cause)
+window_create(const char *name, int argc, char **argv, const char *path,
+    const char *shell, const char *cwd, struct environ *env,
+    struct termios *tio, u_int sx, u_int sy, u_int hlimit, char **cause)
 {
 	struct window		*w;
 	struct window_pane	*wp;
@@ -317,7 +331,8 @@ window_create(const char *name, const char *cmd, const char *shell,
 	wp = window_add_pane(w, hlimit);
 	layout_init(w, wp);
 
-	if (window_pane_spawn(wp, cmd, shell, cwd, env, tio, cause) != 0) {
+	if (window_pane_spawn(wp, argc, argv, path, shell, cwd, env, tio,
+	    cause) != 0) {
 		window_destroy(w);
 		return (NULL);
 	}
@@ -325,7 +340,7 @@ window_create(const char *name, const char *cmd, const char *shell,
 	w->active = TAILQ_FIRST(&w->panes);
 	if (name != NULL) {
 		w->name = xstrdup(name);
-		options_set_number(&w->options, "automatic-rename", 0);
+		options_set_number(w->options, "automatic-rename", 0);
 	} else
 		w->name = default_window_name(w);
 
@@ -335,23 +350,21 @@ window_create(const char *name, const char *cmd, const char *shell,
 void
 window_destroy(struct window *w)
 {
-	u_int	i;
-
-	window_unzoom(w);
-
-	if (window_index(w, &i) != 0)
-		fatalx("index not found");
-	ARRAY_SET(&windows, i, NULL);
-	while (!ARRAY_EMPTY(&windows) && ARRAY_LAST(&windows) == NULL)
-		ARRAY_TRUNC(&windows, 1);
+	RB_REMOVE(windows, &windows, w);
 
 	if (w->layout_root != NULL)
-		layout_free(w);
+		layout_free_cell(w->layout_root);
+	if (w->saved_layout_root != NULL)
+		layout_free_cell(w->saved_layout_root);
+	free(w->old_layout);
 
-	if (event_initialized(&w->name_timer))
-		evtimer_del(&w->name_timer);
+	if (event_initialized(&w->name_event))
+		evtimer_del(&w->name_event);
 
-	options_free(&w->options);
+	if (event_initialized(&w->alerts_timer))
+		evtimer_del(&w->alerts_timer);
+
+	options_free(w->options);
 
 	window_destroy_panes(w);
 
@@ -385,7 +398,9 @@ window_set_name(struct window *w, const char *new_name)
 	free(w->name);
 	w->name = xstrdup(new_name);
 	notify_window_renamed(w);
+#ifdef TMATE
 	tmate_sync_layout();
+#endif
 }
 
 void
@@ -395,11 +410,23 @@ window_resize(struct window *w, u_int sx, u_int sy)
 	w->sy = sy;
 }
 
-void
+int
+window_has_pane(struct window *w, struct window_pane *wp)
+{
+	struct window_pane	*wp1;
+
+	TAILQ_FOREACH(wp1, &w->panes, entry) {
+		if (wp1 == wp)
+			return (1);
+	}
+	return (0);
+}
+
+int
 window_set_active_pane(struct window *w, struct window_pane *wp)
 {
 	if (wp == w->active)
-		return;
+		return (0);
 	w->last = w->active;
 	w->active = wp;
 	while (!window_pane_visible(w->active)) {
@@ -407,8 +434,35 @@ window_set_active_pane(struct window *w, struct window_pane *wp)
 		if (w->active == NULL)
 			w->active = TAILQ_LAST(&w->panes, window_panes);
 		if (w->active == wp)
-			return;
+			return (1);
 	}
+	w->active->active_point = next_active_point++;
+	w->active->flags |= PANE_CHANGED;
+	return (1);
+}
+
+void
+window_redraw_active_switch(struct window *w, struct window_pane *wp)
+{
+	const struct grid_cell	*agc, *wgc;
+
+	if (wp == w->active)
+		return;
+
+	/*
+	 * If window-style and window-active-style are the same, we don't need
+	 * to redraw panes when switching active panes. Otherwise, if the
+	 * active or inactive pane do not have a custom style, they will need
+	 * to be redrawn.
+	 */
+	agc = options_get_style(w->options, "window-active-style");
+	wgc = options_get_style(w->options, "window-style");
+	if (style_equal(agc, wgc))
+		return;
+	if (style_equal(&grid_default_cell, &w->active->colgc))
+		w->active->flags |= PANE_REDRAW;
+	if (style_equal(&grid_default_cell, &wp->colgc))
+		wp->flags |= PANE_REDRAW;
 }
 
 struct window_pane *
@@ -428,16 +482,6 @@ window_get_active_at(struct window *w, u_int x, u_int y)
 	return (NULL);
 }
 
-void
-window_set_active_at(struct window *w, u_int x, u_int y)
-{
-	struct window_pane	*wp;
-
-	wp = window_get_active_at(w, x, y);
-	if (wp != NULL && wp != w->active)
-		window_set_active_pane(w, wp);
-}
-
 struct window_pane *
 window_find_string(struct window *w, const char *s)
 {
@@ -498,6 +542,7 @@ window_zoom(struct window_pane *wp)
 	w->saved_layout_root = w->layout_root;
 	layout_init(w, wp);
 	w->flags |= WINDOW_ZOOMED;
+	notify_window_layout_changed(w);
 
 	return (0);
 }
@@ -513,12 +558,14 @@ window_unzoom(struct window *w)
 	w->flags &= ~WINDOW_ZOOMED;
 	layout_free(w);
 	w->layout_root = w->saved_layout_root;
+	w->saved_layout_root = NULL;
 
 	TAILQ_FOREACH(wp, &w->panes, entry) {
 		wp->layout_cell = wp->saved_layout_cell;
 		wp->saved_layout_cell = NULL;
 	}
 	layout_fix_panes(w, w->sx, w->sy);
+	notify_window_layout_changed(w);
 
 	return (0);
 }
@@ -537,8 +584,11 @@ window_add_pane(struct window *w, u_int hlimit)
 }
 
 void
-window_remove_pane(struct window *w, struct window_pane *wp)
+window_lost_pane(struct window *w, struct window_pane *wp)
 {
+	if (wp == marked_pane.wp)
+		server_clear_marked();
+
 	if (wp == w->active) {
 		w->active = w->last;
 		w->last = NULL;
@@ -547,8 +597,16 @@ window_remove_pane(struct window *w, struct window_pane *wp)
 			if (w->active == NULL)
 				w->active = TAILQ_NEXT(wp, entry);
 		}
+		if (w->active != NULL)
+			w->active->flags |= PANE_CHANGED;
 	} else if (wp == w->last)
 		w->last = NULL;
+}
+
+void
+window_remove_pane(struct window *w, struct window_pane *wp)
+{
+	window_lost_pane(w, wp);
 
 	TAILQ_REMOVE(&w->panes, wp, entry);
 	window_pane_destroy(wp);
@@ -560,7 +618,7 @@ window_pane_at_index(struct window *w, u_int idx)
 	struct window_pane	*wp;
 	u_int			 n;
 
-	n = options_get_number(&w->options, "pane-base-index");
+	n = options_get_number(w->options, "pane-base-index");
 	TAILQ_FOREACH(wp, &w->panes, entry) {
 		if (n == idx)
 			return (wp);
@@ -598,7 +656,7 @@ window_pane_index(struct window_pane *wp, u_int *i)
 	struct window_pane	*wq;
 	struct window		*w = wp->window;
 
-	*i = options_get_number(&w->options, "pane-base-index");
+	*i = options_get_number(w->options, "pane-base-index");
 	TAILQ_FOREACH(wq, &w->panes, entry) {
 		if (wp == wq) {
 			return (0);
@@ -633,11 +691,11 @@ window_destroy_panes(struct window *w)
 	}
 }
 
-/* Return list of printable window flag symbols. No flags is just a space. */
+/* Retuns the printable flags on a window, empty string if no flags set. */
 char *
 window_printable_flags(struct session *s, struct winlink *wl)
 {
-	char	flags[BUFSIZ];
+	char	flags[32];
 	int	pos;
 
 	pos = 0;
@@ -645,23 +703,35 @@ window_printable_flags(struct session *s, struct winlink *wl)
 		flags[pos++] = '#';
 	if (wl->flags & WINLINK_BELL)
 		flags[pos++] = '!';
-	if (wl->flags & WINLINK_CONTENT)
-		flags[pos++] = '+';
 	if (wl->flags & WINLINK_SILENCE)
 		flags[pos++] = '~';
 	if (wl == s->curw)
 		flags[pos++] = '*';
 	if (wl == TAILQ_FIRST(&s->lastw))
 		flags[pos++] = '-';
+	if (server_check_marked() && wl == marked_pane.wl)
+		flags[pos++] = 'M';
 	if (wl->window->flags & WINDOW_ZOOMED)
 		flags[pos++] = 'Z';
-	if (pos == 0)
-		flags[pos++] = ' ';
 	flags[pos] = '\0';
 	return (xstrdup(flags));
 }
 
-/* Find pane in global tree by id. */
+struct window_pane *
+window_pane_find_by_id_str(const char *s)
+{
+	const char	*errstr;
+	u_int		 id;
+
+	if (*s != '%')
+		return (NULL);
+
+	id = strtonum(s + 1, 0, UINT_MAX, &errstr);
+	if (errstr != NULL)
+		return (NULL);
+	return (window_pane_find_by_id(id));
+}
+
 struct window_pane *
 window_pane_find_by_id(u_int id)
 {
@@ -675,6 +745,7 @@ struct window_pane *
 window_pane_create(struct window *w, u_int sx, u_int sy, u_int hlimit)
 {
 	struct window_pane	*wp;
+	char			 host[HOST_NAME_MAX + 1];
 
 	wp = xcalloc(1, sizeof *wp);
 	wp->window = w;
@@ -682,7 +753,8 @@ window_pane_create(struct window *w, u_int sx, u_int sy, u_int hlimit)
 	wp->id = next_window_pane_id++;
 	RB_INSERT(window_pane_tree, &all_window_panes, wp);
 
-	wp->cmd = NULL;
+	wp->argc = 0;
+	wp->argv = NULL;
 	wp->shell = NULL;
 	wp->cwd = NULL;
 
@@ -703,13 +775,20 @@ window_pane_create(struct window *w, u_int sx, u_int sy, u_int hlimit)
 	wp->pipe_off = 0;
 	wp->pipe_event = NULL;
 
+#ifdef TMATE
 	wp->tmate_off = 0;
+#endif
 
 	wp->saved_grid = NULL;
 
+	memcpy(&wp->colgc, &grid_default_cell, sizeof wp->colgc);
+
 	screen_init(&wp->base, sx, sy, hlimit);
 	wp->screen = &wp->base;
 
+	if (gethostname(host, sizeof host) == 0)
+		screen_set_title(&wp->base, host);
+
 	input_init(wp);
 
 	return (wp);
@@ -720,10 +799,13 @@ window_pane_destroy(struct window_pane *wp)
 {
 	window_pane_reset_mode(wp);
 
-	if (event_initialized(&wp->changes_timer))
-		evtimer_del(&wp->changes_timer);
+	if (event_initialized(&wp->timer))
+		evtimer_del(&wp->timer);
 
 	if (wp->fd != -1) {
+#ifdef HAVE_UTEMPTER
+		utempter_remove_record(wp->fd);
+#endif
 		bufferevent_free(wp->event);
 		close(wp->fd);
 	}
@@ -741,39 +823,48 @@ window_pane_destroy(struct window_pane *wp)
 
 	RB_REMOVE(window_pane_tree, &all_window_panes, wp);
 
-	free(wp->cwd);
+	free((void *)wp->cwd);
 	free(wp->shell);
-	free(wp->cmd);
+	cmd_free_argv(wp->argc, wp->argv);
 	free(wp);
 }
 
 int
-window_pane_spawn(struct window_pane *wp, const char *cmd, const char *shell,
-    const char *cwd, struct environ *env, struct termios *tio, char **cause)
+window_pane_spawn(struct window_pane *wp, int argc, char **argv,
+    const char *path, const char *shell, const char *cwd, struct environ *env,
+    struct termios *tio, char **cause)
 {
 	struct winsize	 ws;
-	char		*argv0, paneid[16];
-	const char	*ptr;
+	char		*argv0, *cmd, **argvp;
+	const char	*ptr, *first, *home;
 	struct termios	 tio2;
+#ifdef HAVE_UTEMPTER
+	char		 s[32];
+#endif
+	int		 i;
 
 	if (wp->fd != -1) {
 		bufferevent_free(wp->event);
 		close(wp->fd);
 	}
-	if (cmd != NULL) {
-		free(wp->cmd);
-		wp->cmd = xstrdup(cmd);
+	if (argc > 0) {
+		cmd_free_argv(wp->argc, wp->argv);
+		wp->argc = argc;
+		wp->argv = cmd_copy_argv(argc, argv);
 	}
 	if (shell != NULL) {
 		free(wp->shell);
 		wp->shell = xstrdup(shell);
 	}
 	if (cwd != NULL) {
-		free(wp->cwd);
+		free((void *)wp->cwd);
 		wp->cwd = xstrdup(cwd);
 	}
 
-	log_debug("spawn: %s -- %s", wp->shell, wp->cmd);
+	cmd = cmd_stringify_argv(wp->argc, wp->argv);
+	log_debug("spawn: %s -- %s", wp->shell, cmd);
+	for (i = 0; i < wp->argc; i++)
+		log_debug("spawn: argv[%d] = %s", i, wp->argv[i]);
 
 	memset(&ws, 0, sizeof ws);
 	ws.ws_col = screen_size_x(&wp->base);
@@ -783,10 +874,13 @@ window_pane_spawn(struct window_pane *wp, const char *cmd, const char *shell,
 	case -1:
 		wp->fd = -1;
 		xasprintf(cause, "%s: %s", cmd, strerror(errno));
+		free(cmd);
 		return (-1);
 	case 0:
-		if (chdir(wp->cwd) != 0)
-			chdir("/");
+		if (chdir(wp->cwd) != 0) {
+			if ((home = find_home()) == NULL || chdir(home) != 0)
+				chdir("/");
+		}
 
 		if (tcgetattr(STDIN_FILENO, &tio2) != 0)
 			fatal("tcgetattr failed");
@@ -794,16 +888,16 @@ window_pane_spawn(struct window_pane *wp, const char *cmd, const char *shell,
 			memcpy(tio2.c_cc, tio->c_cc, sizeof tio2.c_cc);
 		tio2.c_cc[VERASE] = '\177';
 #ifdef IUTF8
-		if (options_get_number(&wp->window->options, "utf8"))
-			tio2.c_iflag |= IUTF8;
+		tio2.c_iflag |= IUTF8;
 #endif
 		if (tcsetattr(STDIN_FILENO, TCSANOW, &tio2) != 0)
 			fatal("tcgetattr failed");
 
 		closefrom(STDERR_FILENO + 1);
 
-		xsnprintf(paneid, sizeof paneid, "%%%u", wp->id);
-		environ_set(env, "TMUX_PANE", paneid);
+		if (path != NULL)
+			environ_set(env, "PATH", "%s", path);
+		environ_set(env, "TMUX_PANE", "%%%u", wp->id);
 		environ_push(env);
 
 		clear_signals(1);
@@ -812,115 +906,122 @@ window_pane_spawn(struct window_pane *wp, const char *cmd, const char *shell,
 		setenv("SHELL", wp->shell, 1);
 		ptr = strrchr(wp->shell, '/');
 
-		if (*wp->cmd != '\0') {
-			/* Use the command. */
+		/*
+		 * If given one argument, assume it should be passed to sh -c;
+		 * with more than one argument, use execvp(). If there is no
+		 * arguments, create a login shell.
+		 */
+		if (wp->argc > 0) {
+			if (wp->argc != 1) {
+				/* Copy to ensure argv ends in NULL. */
+				argvp = cmd_copy_argv(wp->argc, wp->argv);
+				execvp(argvp[0], argvp);
+				fatal("execvp failed");
+			}
+			first = wp->argv[0];
+
 			if (ptr != NULL && *(ptr + 1) != '\0')
 				xasprintf(&argv0, "%s", ptr + 1);
 			else
 				xasprintf(&argv0, "%s", wp->shell);
-			execl(wp->shell, argv0, "-c", wp->cmd, (char *) NULL);
+			execl(wp->shell, argv0, "-c", first, (char *)NULL);
 			fatal("execl failed");
 		}
-
-		/* No command; fork a login shell. */
 		if (ptr != NULL && *(ptr + 1) != '\0')
 			xasprintf(&argv0, "-%s", ptr + 1);
 		else
 			xasprintf(&argv0, "-%s", wp->shell);
-		execl(wp->shell, argv0, (char *) NULL);
+		execl(wp->shell, argv0, (char *)NULL);
 		fatal("execl failed");
 	}
 
+#ifdef HAVE_UTEMPTER
+	xsnprintf(s, sizeof s, "tmux(%lu).%%%u", (long) getpid(), wp->id);
+	utempter_add_record(wp->fd, s);
+	kill(getpid(), SIGCHLD);
+#endif
+
 	setblocking(wp->fd, 0);
 
-	wp->event = bufferevent_new(wp->fd,
-	    window_pane_read_callback, NULL, window_pane_error_callback, wp);
+	wp->event = bufferevent_new(wp->fd, window_pane_read_callback, NULL,
+	    window_pane_error_callback, wp);
+
+	bufferevent_setwatermark(wp->event, EV_READ, 0, READ_SIZE);
 	bufferevent_enable(wp->event, EV_READ|EV_WRITE);
 
+	free(cmd);
 	return (0);
 }
 
 void
-window_pane_timer_start(struct window_pane *wp)
+window_pane_timer_callback(__unused int fd, __unused short events, void *data)
 {
-	struct timeval	tv;
-
-	tv.tv_sec = 0;
-	tv.tv_usec = 1000;
-
-	evtimer_del(&wp->changes_timer);
-	evtimer_set(&wp->changes_timer, window_pane_timer_callback, wp);
-	evtimer_add(&wp->changes_timer, &tv);
+	window_pane_read_callback(NULL, data);
 }
 
 void
-window_pane_timer_callback(unused int fd, unused short events, void *data)
+window_pane_read_callback(__unused struct bufferevent *bufev, void *data)
 {
 	struct window_pane	*wp = data;
-	struct window		*w = wp->window;
-	u_int			 interval, trigger;
+	struct evbuffer		*evb = wp->event->input;
+	char			*new_data;
+	size_t			 new_size, available;
+	struct client		*c;
+	struct timeval		 tv;
 
-	interval = options_get_number(&w->options, "c0-change-interval");
-	trigger = options_get_number(&w->options, "c0-change-trigger");
+	if (event_initialized(&wp->timer))
+		evtimer_del(&wp->timer);
 
-	if (wp->changes_redraw++ == interval) {
-		wp->flags |= PANE_REDRAW;
-		wp->changes_redraw = 0;
+	log_debug("%%%u has %zu bytes", wp->id, EVBUFFER_LENGTH(evb));
 
+	TAILQ_FOREACH(c, &clients, entry) {
+		if (!tty_client_ready(c, wp))
+			continue;
+
+		available = EVBUFFER_LENGTH(c->tty.event->output);
+		if (available > READ_BACKOFF)
+			goto start_timer;
 	}
 
-	if (trigger == 0 || wp->changes < trigger) {
-		wp->flags |= PANE_REDRAW;
-		wp->flags &= ~PANE_DROP;
-	} else
-		window_pane_timer_start(wp);
-	wp->changes = 0;
-}
-
-void
-window_pane_read_callback(unused struct bufferevent *bufev, void *data)
-{
-	struct window_pane     *wp = data;
-	char   		       *new_data;
-	size_t			new_size;
-
-	new_size = EVBUFFER_LENGTH(wp->event->input) - wp->pipe_off;
+	new_size = EVBUFFER_LENGTH(evb) - wp->pipe_off;
 	if (wp->pipe_fd != -1 && new_size > 0) {
-		/* FIXME tmux:
-		 * - new_data = EVBUFFER_DATA(wp->event->input);
-		 * + new_data = EVBUFFER_DATA(wp->event->input) + wp->pipe_off;
-		 * also, can the buffer be too small?
-		 */
-		new_data = EVBUFFER_DATA(wp->event->input);
+		new_data = EVBUFFER_DATA(evb) + wp->pipe_off;
 		bufferevent_write(wp->pipe_event, new_data, new_size);
 	}
 
+#ifdef TMATE
 	new_size = EVBUFFER_LENGTH(wp->event->input) - wp->tmate_off;
 	new_data = EVBUFFER_DATA(wp->event->input) + wp->tmate_off;
 	if (new_size > 0)
 		tmate_pty_data(wp, new_data, new_size);
+#endif
 
 	input_parse(wp);
 
-	wp->pipe_off = EVBUFFER_LENGTH(wp->event->input);
-	wp->tmate_off = EVBUFFER_LENGTH(wp->event->input);
+	wp->pipe_off = EVBUFFER_LENGTH(evb);
+#ifdef TMATE
+	wp->tmate_off = EVBUFFER_LENGTH(evb);
+#endif
+	return;
 
-	/*
-	 * If we get here, we're not outputting anymore, so set the silence
-	 * flag on the window.
-	 */
-	wp->window->flags |= WINDOW_SILENCE;
-	if (gettimeofday(&wp->window->silence_timer, NULL) != 0)
-		fatal("gettimeofday failed.");
+start_timer:
+	log_debug("%%%u backing off (%s %zu > %d)", wp->id, c->ttyname,
+	    available, READ_BACKOFF);
+
+	tv.tv_sec = 0;
+	tv.tv_usec = READ_TIME;
+
+	evtimer_set(&wp->timer, window_pane_timer_callback, wp);
+	evtimer_add(&wp->timer, &tv);
 }
 
 void
-window_pane_error_callback(
-    unused struct bufferevent *bufev, unused short what, void *data)
+window_pane_error_callback(__unused struct bufferevent *bufev,
+    __unused short what, void *data)
 {
 	struct window_pane *wp = data;
 
-	server_destroy_pane(wp);
+	server_destroy_pane(wp, 1);
 }
 
 void
@@ -951,7 +1052,7 @@ window_pane_alternate_on(struct window_pane *wp, struct grid_cell *gc,
 
 	if (wp->saved_grid != NULL)
 		return;
-	if (!options_get_number(&wp->window->options, "alternate-screen"))
+	if (!options_get_number(wp->window->options, "alternate-screen"))
 		return;
 	sx = screen_size_x(s);
 	sy = screen_size_y(s);
@@ -981,7 +1082,7 @@ window_pane_alternate_off(struct window_pane *wp, struct grid_cell *gc,
 
 	if (wp->saved_grid == NULL)
 		return;
-	if (!options_get_number(&wp->window->options, "alternate-screen"))
+	if (!options_get_number(wp->window->options, "alternate-screen"))
 		return;
 	sx = screen_size_x(s);
 	sy = screen_size_y(s);
@@ -1030,7 +1131,9 @@ window_pane_set_mode(struct window_pane *wp, const struct window_mode *mode)
 
 	if ((s = wp->mode->init(wp)) != NULL)
 		wp->screen = s;
-	wp->flags |= PANE_REDRAW;
+	wp->flags |= (PANE_REDRAW|PANE_CHANGED);
+
+	server_status_window(wp->window);
 	return (0);
 }
 
@@ -1044,60 +1147,49 @@ window_pane_reset_mode(struct window_pane *wp)
 	wp->mode = NULL;
 
 	wp->screen = &wp->base;
-	wp->flags |= PANE_REDRAW;
+	wp->flags |= (PANE_REDRAW|PANE_CHANGED);
 
+	server_status_window(wp->window);
+
+#ifdef TMATE
 	tmate_sync_copy_mode(wp);
+#endif
 }
 
 void
-window_pane_key(struct window_pane *wp, struct session *sess, int key)
+window_pane_key(struct window_pane *wp, struct client *c, struct session *s,
+    key_code key, struct mouse_event *m)
 {
 	struct window_pane	*wp2;
 
-	if (!window_pane_visible(wp))
+	if (KEYC_IS_MOUSE(key) && m == NULL)
 		return;
 
 	if (wp->mode != NULL) {
 		if (wp->mode->key != NULL)
-			wp->mode->key(wp, sess, key);
+			wp->mode->key(wp, c, s, key, m);
 		return;
 	}
 
-	if (wp->fd == -1)
+	if (wp->fd == -1 || wp->flags & PANE_INPUTOFF)
 		return;
-	input_key(wp, key);
-	if (options_get_number(&wp->window->options, "synchronize-panes")) {
+
+	input_key(wp, key, m);
+
+	if (KEYC_IS_MOUSE(key))
+		return;
+	if (options_get_number(wp->window->options, "synchronize-panes")) {
 		TAILQ_FOREACH(wp2, &wp->window->panes, entry) {
 			if (wp2 == wp || wp2->mode != NULL)
 				continue;
-			if (wp2->fd != -1 && window_pane_visible(wp2))
-				input_key(wp2, key);
+			if (wp2->fd == -1 || wp2->flags & PANE_INPUTOFF)
+				continue;
+			if (window_pane_visible(wp2))
+				input_key(wp2, key, NULL);
 		}
 	}
 }
 
-void
-window_pane_mouse(
-    struct window_pane *wp, struct session *sess, struct mouse_event *m)
-{
-	if (!window_pane_visible(wp))
-		return;
-
-	if (m->x < wp->xoff || m->x >= wp->xoff + wp->sx)
-		return;
-	if (m->y < wp->yoff || m->y >= wp->yoff + wp->sy)
-		return;
-	m->x -= wp->xoff;
-	m->y -= wp->yoff;
-
-	if (wp->mode != NULL) {
-		if (wp->mode->mouse != NULL &&
-		    options_get_number(&wp->window->options, "mode-mouse"))
-			wp->mode->mouse(wp, sess, m);
-	} else if (wp->fd != -1)
-		input_mouse(wp, sess, m);
-}
-
 int
 window_pane_visible(struct window_pane *wp)
 {
@@ -1113,7 +1205,8 @@ window_pane_visible(struct window_pane *wp)
 }
 
 char *
-window_pane_search(struct window_pane *wp, const char *searchstr, u_int *lineno)
+window_pane_search(struct window_pane *wp, const char *searchstr,
+    u_int *lineno)
 {
 	struct screen	*s = &wp->base;
 	char		*newsearchstr, *line, *msg;
@@ -1137,150 +1230,254 @@ window_pane_search(struct window_pane *wp, const char *searchstr, u_int *lineno)
 	return (msg);
 }
 
-/* Find the pane directly above another. */
+/* Get MRU pane from a list. */
+struct window_pane *
+window_pane_choose_best(struct window_pane **list, u_int size)
+{
+	struct window_pane	*next, *best;
+	u_int			 i;
+
+	if (size == 0)
+		return (NULL);
+
+	best = list[0];
+	for (i = 1; i < size; i++) {
+		next = list[i];
+		if (next->active_point > best->active_point)
+			best = next;
+	}
+	return (best);
+}
+
+/*
+ * Find the pane directly above another. We build a list of those adjacent to
+ * top edge and then choose the best.
+ */
 struct window_pane *
 window_pane_find_up(struct window_pane *wp)
 {
-	struct window_pane     *wp2;
-	u_int			left, top;
+	struct window_pane	*next, *best, **list;
+	u_int			 edge, left, right, end, size;
+	int			 found;
 
 	if (wp == NULL || !window_pane_visible(wp))
 		return (NULL);
 
-	top = wp->yoff;
-	if (top == 0)
-		top = wp->window->sy + 1;
-	left = wp->xoff;
+	list = NULL;
+	size = 0;
 
-	TAILQ_FOREACH(wp2, &wp->window->panes, entry) {
-		if (!window_pane_visible(wp2))
+	edge = wp->yoff;
+	if (edge == 0)
+		edge = wp->window->sy + 1;
+
+	left = wp->xoff;
+	right = wp->xoff + wp->sx;
+
+	TAILQ_FOREACH(next, &wp->window->panes, entry) {
+		if (next == wp || !window_pane_visible(next))
 			continue;
-		if (wp2->yoff + wp2->sy + 1 != top)
+		if (next->yoff + next->sy + 1 != edge)
 			continue;
-		if (left >= wp2->xoff && left <= wp2->xoff + wp2->sx)
-			return (wp2);
+		end = next->xoff + next->sx - 1;
+
+		found = 0;
+		if (next->xoff < left && end > right)
+			found = 1;
+		else if (next->xoff >= left && next->xoff <= right)
+			found = 1;
+		else if (end >= left && end <= right)
+			found = 1;
+		if (!found)
+			continue;
+		list = xreallocarray(list, size + 1, sizeof *list);
+		list[size++] = next;
 	}
-	return (NULL);
+
+	best = window_pane_choose_best(list, size);
+	free(list);
+	return (best);
 }
 
 /* Find the pane directly below another. */
 struct window_pane *
 window_pane_find_down(struct window_pane *wp)
 {
-	struct window_pane     *wp2;
-	u_int			left, bottom;
+	struct window_pane	*next, *best, **list;
+	u_int			 edge, left, right, end, size;
+	int			 found;
 
 	if (wp == NULL || !window_pane_visible(wp))
 		return (NULL);
 
-	bottom = wp->yoff + wp->sy + 1;
-	if (bottom >= wp->window->sy)
-		bottom = 0;
-	left = wp->xoff;
+	list = NULL;
+	size = 0;
 
-	TAILQ_FOREACH(wp2, &wp->window->panes, entry) {
-		if (!window_pane_visible(wp2))
+	edge = wp->yoff + wp->sy + 1;
+	if (edge >= wp->window->sy)
+		edge = 0;
+
+	left = wp->xoff;
+	right = wp->xoff + wp->sx;
+
+	TAILQ_FOREACH(next, &wp->window->panes, entry) {
+		if (next == wp || !window_pane_visible(next))
 			continue;
-		if (wp2->yoff != bottom)
+		if (next->yoff != edge)
 			continue;
-		if (left >= wp2->xoff && left <= wp2->xoff + wp2->sx)
-			return (wp2);
+		end = next->xoff + next->sx - 1;
+
+		found = 0;
+		if (next->xoff < left && end > right)
+			found = 1;
+		else if (next->xoff >= left && next->xoff <= right)
+			found = 1;
+		else if (end >= left && end <= right)
+			found = 1;
+		if (!found)
+			continue;
+		list = xreallocarray(list, size + 1, sizeof *list);
+		list[size++] = next;
 	}
-	return (NULL);
+
+	best = window_pane_choose_best(list, size);
+	free(list);
+	return (best);
 }
 
-/*
- * Find the pane directly to the left of another, adjacent to the left side and
- * containing the top edge.
- */
+/* Find the pane directly to the left of another. */
 struct window_pane *
 window_pane_find_left(struct window_pane *wp)
 {
-	struct window_pane     *wp2;
-	u_int			left, top;
+	struct window_pane	*next, *best, **list;
+	u_int			 edge, top, bottom, end, size;
+	int			 found;
 
 	if (wp == NULL || !window_pane_visible(wp))
 		return (NULL);
 
-	left = wp->xoff;
-	if (left == 0)
-		left = wp->window->sx + 1;
-	top = wp->yoff;
+	list = NULL;
+	size = 0;
 
-	TAILQ_FOREACH(wp2, &wp->window->panes, entry) {
-		if (!window_pane_visible(wp2))
+	edge = wp->xoff;
+	if (edge == 0)
+		edge = wp->window->sx + 1;
+
+	top = wp->yoff;
+	bottom = wp->yoff + wp->sy;
+
+	TAILQ_FOREACH(next, &wp->window->panes, entry) {
+		if (next == wp || !window_pane_visible(next))
 			continue;
-		if (wp2->xoff + wp2->sx + 1 != left)
+		if (next->xoff + next->sx + 1 != edge)
 			continue;
-		if (top >= wp2->yoff && top <= wp2->yoff + wp2->sy)
-			return (wp2);
+		end = next->yoff + next->sy - 1;
+
+		found = 0;
+		if (next->yoff < top && end > bottom)
+			found = 1;
+		else if (next->yoff >= top && next->yoff <= bottom)
+			found = 1;
+		else if (end >= top && end <= bottom)
+			found = 1;
+		if (!found)
+			continue;
+		list = xreallocarray(list, size + 1, sizeof *list);
+		list[size++] = next;
 	}
-	return (NULL);
+
+	best = window_pane_choose_best(list, size);
+	free(list);
+	return (best);
 }
 
-/*
- * Find the pane directly to the right of another, that is adjacent to the
- * right edge and including the top edge.
- */
+/* Find the pane directly to the right of another. */
 struct window_pane *
 window_pane_find_right(struct window_pane *wp)
 {
-	struct window_pane     *wp2;
-	u_int			right, top;
+	struct window_pane	*next, *best, **list;
+	u_int			 edge, top, bottom, end, size;
+	int			 found;
 
 	if (wp == NULL || !window_pane_visible(wp))
 		return (NULL);
 
-	right = wp->xoff + wp->sx + 1;
-	if (right >= wp->window->sx)
-		right = 0;
-	top = wp->yoff;
+	list = NULL;
+	size = 0;
 
-	TAILQ_FOREACH(wp2, &wp->window->panes, entry) {
-		if (!window_pane_visible(wp2))
+	edge = wp->xoff + wp->sx + 1;
+	if (edge >= wp->window->sx)
+		edge = 0;
+
+	top = wp->yoff;
+	bottom = wp->yoff + wp->sy;
+
+	TAILQ_FOREACH(next, &wp->window->panes, entry) {
+		if (next == wp || !window_pane_visible(next))
 			continue;
-		if (wp2->xoff != right)
+		if (next->xoff != edge)
 			continue;
-		if (top >= wp2->yoff && top <= wp2->yoff + wp2->sy)
-			return (wp2);
+		end = next->yoff + next->sy - 1;
+
+		found = 0;
+		if (next->yoff < top && end > bottom)
+			found = 1;
+		else if (next->yoff >= top && next->yoff <= bottom)
+			found = 1;
+		else if (end >= top && end <= bottom)
+			found = 1;
+		if (!found)
+			continue;
+		list = xreallocarray(list, size + 1, sizeof *list);
+		list[size++] = next;
 	}
-	return (NULL);
+
+	best = window_pane_choose_best(list, size);
+	free(list);
+	return (best);
 }
 
 /* Clear alert flags for a winlink */
 void
 winlink_clear_flags(struct winlink *wl)
 {
-	struct winlink	*wm;
 	struct session	*s;
-	struct window	*w;
-	u_int		 i;
+	struct winlink	*wl_loop;
 
-	for (i = 0; i < ARRAY_LENGTH(&windows); i++) {
-		if ((w = ARRAY_ITEM(&windows, i)) == NULL)
-			continue;
-
-		RB_FOREACH(s, sessions, &sessions) {
-			if ((wm = session_has(s, w)) == NULL)
+	RB_FOREACH(s, sessions, &sessions) {
+		RB_FOREACH(wl_loop, winlinks, &s->windows) {
+			if (wl_loop->window != wl->window)
+				continue;
+			if ((wl_loop->flags & WINLINK_ALERTFLAGS) == 0)
 				continue;
 
-			if (wm->window != wl->window)
-				continue;
-			if ((wm->flags & WINLINK_ALERTFLAGS) == 0)
-				continue;
-
-			wm->flags &= ~WINLINK_ALERTFLAGS;
+			wl_loop->flags &= ~WINLINK_ALERTFLAGS;
+			wl_loop->window->flags &= ~WINDOW_ALERTFLAGS;
 			server_status_session(s);
 		}
 	}
 }
 
-/* Set the grid_cell with fg/bg/attr information when window is in a mode. */
-void
-window_mode_attrs(struct grid_cell *gc, struct options *oo)
+int
+winlink_shuffle_up(struct session *s, struct winlink *wl)
 {
-	memcpy(gc, &grid_default_cell, sizeof *gc);
-	colour_set_fg(gc, options_get_number(oo, "mode-fg"));
-	colour_set_bg(gc, options_get_number(oo, "mode-bg"));
-	gc->attr |= options_get_number(oo, "mode-attr");
+	int	 idx, last;
+
+	idx = wl->idx + 1;
+
+	/* Find the next free index. */
+	for (last = idx; last < INT_MAX; last++) {
+		if (winlink_find_by_index(&s->windows, last) == NULL)
+			break;
+	}
+	if (last == INT_MAX)
+		return (-1);
+
+	/* Move everything from last - 1 to idx up a bit. */
+	for (; last > idx; last--) {
+		wl = winlink_find_by_index(&s->windows, last - 1);
+		server_link_window(s, wl, s, last, 0, 0, NULL);
+		server_unlink_window(s, wl);
+	}
+
+	return (idx);
 }
